import functools
import typing
from contextlib import contextmanager
from copy import deepcopy
from functools import partial
import logging
import sqlalchemy as sa
import sqlalchemy.exc
import sqlalchemy.orm.session as orms
from box import Box
from geoalchemy2.shape import to_shape
from sqlalchemy.engine.url import URL
from sqlalchemy.orm import sessionmaker

from . import sqla_orms
from .. import logger
from ..common import utils

def get_engine(db_configs: dict,create_if_not_exists=False,drivername='postgres',**connection_kwargs):
	db_configs=db_configs.copy()
	# database=
	database=connection_kwargs.pop('database') if connection_kwargs.get('database') else db_configs.pop('database')
	# if connection_kwargs.get('database'): database=connection_kwargs.pop('database')
	# elif db_configs.get('database'): database=db_configs.pop('database')
	engine_creator=lambda db:sa.create_engine(URL(**dict(dict(db_configs,drivername=drivername,database=db),
	                                                     **connection_kwargs)),echo=False)

	logger.trace(logging.INFO,skimpy=True)

	def __create_database():
		eng=engine_creator('postgres')
		conn=eng.connect()
		conn.execute('commit')
		conn.execute(f'create database {database};')
		conn.close()

	if create_if_not_exists:
		try:
			engine=engine_creator(database)
			engine.connect()
		except sa.exc.OperationalError:
			logger.info(f"Database '%s' does not exist.",database)
			__create_database()
			engine=engine_creator(database)
	else:
		engine=engine_creator(database)
	engine.connect().execute("create extension if not exists postgis;"
	                         "create extension if not exists postgis_topology;")
	# simple_logger.info('engine: %s',engine)
	return engine

def get_session(engine,autoflush=True,autocommit=False,expire_on_commit=True,info=None,**kwargs) -> orms.Session:
	session: orms.Session=sessionmaker(bind=engine,autoflush=autoflush,autocommit=autocommit,expire_on_commit=expire_on_commit,
	                                   info=info,**kwargs)()
	return session

@contextmanager
def session_context(engine,autoflush=True,autocommit=False,expire_on_commit=True,info=None,**kwargs):
	session=get_session(engine,autoflush=autoflush,autocommit=autocommit,expire_on_commit=expire_on_commit,info=info,**kwargs)
	try:
		yield session
		session.commit()
	except Exception as e:
		logger.exception(str(e))
		session.rollback()
		raise
	finally:
		session.close()

class QueryItem(sqla_orms.Reconnectable):
	_protected_keys=['info',*['_QueryItem'+x for x in ['__item','__engine_creator','__obj_class','__primary_keys']]]

	def __init__(self,obj: sqla_orms.Base,engine_creator,pkeys):
		self.__item=Box(deepcopy(obj.__dict__))
		self.__obj_class=obj.__class__
		self.__engine_creator=engine_creator
		self.__primary_keys=tuple([x for x in pkeys] if type(pkeys[0]) is str else [x.key for x in pkeys])

	@property
	def primary_vals(self):
		return tuple(getattr(self.__item,x) for x in self.__primary_keys)

	@property
	def info(self):
		return sqla_orms.get_info(self.__item)

	def reconnect(self,session):
		target_obj=session.query(self.__obj_class).get(self.primary_vals)
		assert target_obj is not None,f"Query on {self.__obj_class} returned None."
		return target_obj

	def get_wkt(self,key='geom'):
		return to_shape(self.__item[key]).wkt

	def __setattr__(self,key,value):
		if key in self._protected_keys:
			self.__dict__[key]=value
		else:
			with session_context(self.__engine_creator()) as temp_session:
				setattr(self.reconnect(temp_session),key,value)
				self.__item[key]=value

	def __repr__(self):
		return f"QueryItem: {utils.reprer(self.__item)}"

	def __getattr__(self,item):
		if item in self._protected_keys:
			return self.__dict__[item]
		return self.__item[item]

	def __getstate__(self):
		return self.__dict__

	def __setstate__(self,state):
		for k,v in state.items():
			setattr(self,k,v)

class Manager(object):
	_protected_keys=['_Manager'+x for x in ['__engine_creator','__engine','__session','__fire_session',
	                                        '__box','__query_item_creator','__temp_ident']]

	def __init__(self,db_configs: dict,create_if_not_exists=False,**connection_kwargs):
		self.__engine_creator=partial(get_engine,db_configs,create_if_not_exists,**connection_kwargs)
		self.__engine=None
		self.__session=None
		self.__query_item_creator=partial(QueryItem,engine_creator=self.__engine_creator)
		self.__box=Box()
		self.__temp_ident=None

	@property
	def engine(self):
		if self.__engine is None:
			self.__engine=self.__engine_creator()
		return self.__engine

	@property
	def session(self):
		if self.__session is None or not self.__session.is_active:
			self.__session=get_session(self.engine)
		return self.__session

	def new_session(self,autoflush=True,autocommit=False,expire_on_commit=True,info=None,**kwargs):
		self.__session=session=get_session(self.engine,autoflush=autoflush,autocommit=autocommit,
		                                   expire_on_commit=expire_on_commit,info=info,**kwargs)
		return session

	@staticmethod
	def commit_and_close(session):
		session.commit()
		session.close()

	def create_table(self,table_class: sqla_orms.Base):
		if self.engine.dialect.has_table(self.engine,table_class.__tablename__):
			logger.info(f"Table '%s' already exists.",table_class.__tablename__)
		else:
			logger.info(f"Creating table '%s'...",table_class.__tablename__)
			table_class.__table__.create(self.engine)
		return self

	@contextmanager
	def session_context(self,**kwargs) -> orms.Session:
		self.__session=session=get_session(self.engine,**kwargs)
		try:
			yield session
			session.commit()
		except Exception as e:
			logger.exception(str(e))
			session.rollback()
			raise
		finally:
			session.close()

	def query_get(self,entity: typing.Type[sqla_orms.KeysBase],*primary_keys):
		assert len(primary_keys)>0,'When querying using query_get, you should provide at least one primary_key.'
		with self.session_context() as session:
			obj=QueryItem(session.query(entity).get(primary_keys),self.__engine_creator,entity.primary_keys)
			return obj

	def get_wrapped(self,obj):
		return QueryItem(obj,self.__engine_creator,obj.primary_keys)

	def session_wrapper(self,new_engine=True,autoflush=True,autocommit=False,expire_on_commit=True,info=None,**kwargs):
		def wrapper_getter(func):
			@functools.wraps(func)
			def wrapper(*args,**_kwargs):
				has_classarg=utils.func_has_classarg(func,args)
				if new_engine:
					self.__engine=self.__engine_creator()
				with self.session_context(autoflush=autoflush,autocommit=autocommit,
				                          expire_on_commit=expire_on_commit,info=info,**kwargs) as session:
					if has_classarg:
						result=func(args[0],session,*args[1:],**_kwargs)
					else:
						result=func(session,*args,**_kwargs)
					return result

			return wrapper

		return wrapper_getter



import datetime
from timeit import default_timer as timer
import wrapt
import logging
import sys
from logging.handlers import RotatingFileHandler

# FORMATTER=logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
# LOG_FILE="lgblkb_tools.log"

# class TheLogger(object):
# 	def __init__(self,logger_name):
# 		pass
#
# def get_console_handler():
# 	console_handler=logging.StreamHandler(sys.stdout)
# 	console_handler.setFormatter(FORMATTER)
# 	return console_handler
#
# def get_file_handler():
# 	file_handler=TimedRotatingFileHandler(LOG_FILE,when='midnight',backupCount=3)
# 	file_handler.setFormatter(FORMATTER)
# 	return file_handler
#
# def get_logger(logger_name):
# 	logger=logging.getLogger(logger_name)
# 	logger.setLevel(logging.DEBUG)  # better to have too much log than not enough
# 	logger.addHandler(get_console_handler())
# 	# logger.addHandler(get_file_handler())
# 	# with this pattern, it's rarely necessary to propagate the error up to parent
# 	logger.propagate=False
# 	return logger
from box import Box
from colorlog import ColoredFormatter
from python_log_indenter import IndentedLoggerAdapter

colored_log_format=f"%(log_color)s%(asctime)s %(levelname)5.5s %(message)s"
# colored_log_format=f"%(asctime)s — %(name)s — %(levelname)s — %(message)s"
base_log_indent_settings=dict(indent_char='| ',spaces=1)
# date_fmt="%m-%d %H:%M:%S"
date_fmt="%H:%M:%S"
default_log_format=colored_log_format.replace(r'%(log_color)s','')

# region log_func_mapper:
log_func_mapper=Box()
log_func_mapper[logging.DEBUG]=lambda _logger:_logger.debug
log_func_mapper[logging.INFO]=lambda _logger:_logger.info
log_func_mapper[logging.WARNING]=lambda _logger:_logger.warning
log_func_mapper[logging.CRITICAL]=lambda _logger:_logger.critical
log_func_mapper[logging.ERROR]=lambda _logger:_logger.error
# endregion

log_sayer=lambda _logger,log_level:log_func_mapper[log_level](_logger)

def generate_handle():
	@wrapt.decorator
	def _handle_generator(wrapped,instance,args,kwargs):
		log_detailer=lambda *_args,level=logging.DEBUG,log_format='',**_kwargs:(level,log_format,_args,_kwargs)
		the_level,the_format,args,kwargs=log_detailer(*args,**kwargs)
		instance.create_handler(wrapped(*args,**kwargs),level=the_level,log_format=the_format)
		instance.info('Log handle %s created at %s',instance.logger.handlers[-1].__class__.__name__,datetime.datetime.now())
		return instance

	return _handle_generator

# noinspection PyUnusedLocal
class TheLogger(IndentedLoggerAdapter):
	def __init__(self,logger_name,level=logging.DEBUG,extra=None,auto_add=True,propagate=False,_logger=None,**kwargs):
		if _logger is None:
			_logger=logging.getLogger(logger_name)
		_logger.setLevel(level)
		super(TheLogger,self).__init__(_logger,extra=extra,auto_add=auto_add,**dict(base_log_indent_settings,**kwargs))
		_logger.propagate=propagate

	def create_handler(self,handler: logging.Handler,level=logging.DEBUG,log_format=''):
		if log_format:
			if type(handler) is logging.StreamHandler: log_formatter_class=ColoredFormatter
			else: log_formatter_class=logging.Formatter
			log_formatter=log_formatter_class(log_format,date_fmt)
		else:
			# if type(handler) is logging.StreamHandler:
			# 	log_formatter=ColoredFormatter(colored_log_format,date_fmt)
			# else:
			log_formatter=logging.Formatter(default_log_format,date_fmt)
		handler.setFormatter(log_formatter)
		handler.setLevel(level=level)
		self.logger.addHandler(handler)
		return self

	@generate_handle()
	def to_rotate(self,log_filepath,maxBytes=2.5e6,backupCount=14,**kwargs):
		return RotatingFileHandler(filename=log_filepath,maxBytes=int(maxBytes),backupCount=backupCount)

	def trace(self,level=logging.DEBUG,skimpy: bool = False,verbose=False):
		log_say=log_sayer(self,level)

		@wrapt.decorator
		def wrapper(wrapped,instance,args,kwargs):
			if skimpy: log_say('Performing "%s"...',wrapped.__name__)
			else: log_say('Running "%s":',wrapped.__name__)
			if verbose:
				logger.debug('args_count: %s',len(args)).a(3)
				for arg in args:
					logger.debug('arg: %s',arg)
				logger.s(3)
				logger.debug('kwargs_count: %s',len(kwargs)).a(3)
				for k,v in kwargs.items():
					logger.debug('%s: %s',k,v)
				logger.s(3)
			start=timer()
			self.add()
			out=wrapped(*args,**kwargs)
			if not skimpy: log_say('Done "%s". Duration: %.3f sec.',wrapped.__name__,timer()-start)
			self.sub()
			return out

		return wrapper

	@generate_handle()
	def to_file(self,log_filepath,**kwargs):
		return logging.FileHandler(log_filepath,**kwargs)

	@generate_handle()
	def to_stream(self,stream=None):
		return logging.StreamHandler(stream or sys.stdout)



logger: TheLogger=TheLogger('lgblkb_logger').to_stream()
logger.info('Log start datetime: %s',datetime.datetime.now())
